from typing import Sequence, Callable, Awaitable, Unpack
from datetime import datetime, timedelta
from sqlmodel import select, Session
from haskellian import iter as I
import chess_pairings as cp
import chess_scraping as cs
from moveread.dfy import Group, Round, Pairings, Game, Tournament, queries
from haskellian import Either, Left, Right
from dslog import Logger

def current_pairings(session: Session, now: datetime | None = None):
  now = now or datetime.now()
  tomorrow = now + timedelta(days=1)
  yesterday = now - timedelta(days=1)
  stmt = select(Tournament, Pairings) \
    .where(Tournament.start_date < tomorrow, yesterday < Tournament.end_date) \
    .join(Pairings, Tournament.tournId == Pairings.tournId) # type: ignore
  return session.exec(stmt).all()

def paired_round(session: Session, **roundId: Unpack[cp.RoundId]):
  """Does the round have pairings?"""
  stmt = queries.select_round_games(**roundId)
  return session.exec(stmt).first() is not None

def finished_round(session: Session, **roundId: Unpack[cp.RoundId]):
  """Does the round have all results? (WARNING: will return true if unpaired! - by vacuity)"""
  stmt = queries.select_round_games(**roundId).where(Game.result == None)
  return session.exec(stmt).first() is None

def current_round(
  session: Session, tournId: str, group: str, *, now: datetime | None = None,
  paired_round = paired_round, finished_round = finished_round
):
  """Determine the current round, whose pairings should be polled"""
  now = now or datetime.now()
  if not (rounds := session.exec(queries.select_rounds(tournId, group)).all()):
    return Left(f'No rounds found for group "{tournId}/{group}"')
  
  if (prev := I.find_last_idx(lambda r: r.start_dtime < now, rounds)) is None:
    return Right(rounds[0].name)
  
  # if next round (which hasn't yet started) is paired, nothing to do
  if paired_round(session, **cp.roundId(tournId, group, rounds[prev+1].name)):
    return Right(None)
  
  # if current round is finished, poll the next round
  if finished_round(session, **cp.roundId(tournId, group, rounds[prev].name)):
    return Right(rounds[prev+1].name)
  
  # otherwise, poll the current round
  return Right(rounds[prev].name)

def update_rounds(session: Session, tnmt: Tournament, group: str, rounds: Sequence[str]) -> Either[str, str]:
  """Update the group's `rounds` list"""
  grp = session.get(Group, (tnmt.tournId, group))
  if grp is None:
    return Left(f'Group "{tnmt.tournId}/{group}" not found')
  elif grp.rounds != rounds:
    og_rounds = [*grp.rounds]
    grp.rounds = rounds
    session.add(grp)
    session.commit()
    return Right(f'Updated group "{tnmt.tournId}/{group}". Rounds {og_rounds} -> {rounds}')
  else:
    return Right(f'Group "{tnmt.tournId}/{group}" already has rounds {rounds}')
  
def update_round_pairings(session: Session, *, tournId: str, group: str, round: str, pairings: cp.RoundPairings):
  """Insert/update pairings of a round"""
  round_games = session.exec(queries.select_round_games(**cp.roundId(tournId, group, round))).all()
  games_idx = cp.GamesMapping[Game].from_pairs([(g.gameId(), g) for g in round_games])
  added: list[cp.GameId] = []
  updated: list[cp.GameId] = []

  for board, pair in pairings.items():
    if pair.tag != 'paired':
      continue
    gid = cp.gameId(tournId, group, round, board)
    if gid in games_idx:
      game = games_idx[gid]
      if game.white != pair.white or game.black != pair.black or game.result != pair.result:
        game.white = pair.white; game.black = pair.black; game.result = pair.result
        updated.append(gid)
    else:
      game = Game(
        tournId=tournId, group=group, round=round, board=board, index=int(board)-1,
        white=pair.white, black=pair.black, result=pair.result
      )
      added.append(gid)
    session.add(game)
  
  session.commit()
  return added, updated
    
  
def update_group_pairings(session: Session, *, tournId: str, group: str, pairings: cp.GroupPairings):
    """Insert/update pairings of a group"""
    group_games = session.exec(queries.select_group_games(tournId, group)).all()
    games_idx = cp.GamesMapping[Game].from_pairs([(g.gameId(), g) for g in group_games])
    added: list[cp.GameId] = []
    updated: list[cp.GameId] = []

    for round, rnd_pairings in pairings.items():
      for board, pair in rnd_pairings.items():
        if pair.tag != 'paired':
          continue
        gid = cp.gameId(tournId, group, round, board)
        if gid in games_idx:
          game = games_idx[gid]
          if game.white != pair.white or game.black != pair.black or game.result != pair.result:
            game.white = pair.white; game.black = pair.black; game.result = pair.result
            updated.append(gid)
        else:
          game = Game(
            tournId=tournId, group=group, round=round, board=board, index=int(board)-1,
            white=pair.white, black=pair.black, result=pair.result
          )
          added.append(gid)
        
        session.add(game)
    
    session.commit()
    return added, updated


async def scrape_all_pairings(
  session: Session, *, now: datetime | None = None, logger: Logger = Logger.click(),
  scrape_group: Callable[[cs.Source], Awaitable[Either[cs.ScrapingError, cp.GroupPairings]]] = cs.scrape_group
):
  """Updates pairings for all ongoing tournaments"""
  pairing_sources = current_pairings(session, now)
  current_tnmts = [tnmt.tournId for tnmt, _ in pairing_sources]
  logger(f'Updating current tournaments: {", ".join(current_tnmts)}')
  for tnmt, src in pairing_sources:
    e = await scrape_group(src.pairings)
    if e.tag == 'left':
      logger(f'Error fetching pairings for "{tnmt.tournId}/{src.group}"', e.value, level='ERROR')
      continue
    pairings = e.value
    rounds = [str(i+1) for i in range(len(pairings))]
    r = update_rounds(session, tnmt, src.group, rounds)
    logger(r.value, level='DEBUG' if r.tag == 'right' else 'ERROR')
    added, updated = update_group_pairings(session, tournId=tnmt.tournId, group=src.group, pairings=pairings)
    logger(f'Updated pairings for "{tnmt.tournId}/{src.group}", added {len(added)} and updated {len(updated)} games', level='DEBUG')


from typing import Sequence, Literal, TypeAlias, Unpack, Any, Iterable, Coroutine
import asyncio
from uuid import uuid4
from sqlalchemy import Engine
from sqlmodel import Session, select
from fastapi import UploadFile
from haskellian import either as E, Left, Either, promise as P, Iter
import pure_cv as vc
from kv import KV, InexistentItem
import chess_pairings as cp
from .pgns import export_all
from .util import TimedIterableCache
from moveread.dfy import Game, FrontendGame, Token, Group, Image, Tournament, FrontendPGN, queries

ImgExtension: TypeAlias = Literal['.jpg', '.png', '.webp']
MimeType: TypeAlias = str

def stringify(tournId: str, group: str, round: str, board: str) -> str:
  return f'{tournId}/{group}/{round}/{board}'

def safe_extension(file: UploadFile) -> str | None:
  if file.filename and len(parts := file.filename.split('.')) == 2:
    return parts[1]
  
def descale2jpg(img: bytes, height: int):
  mat = vc.decode(img)
  return vc.encode(vc.descale_h(mat, height), format='.jpg')
  
class SDK:

  def __init__(self, images: KV[bytes], engine: Engine, *, cache_ttl: float = 30, cache_max_entries: int = 1000):
    self._images = images
    self._engine = engine
    self.pgns_cache = TimedIterableCache(self._round_pgn, ttl_secs=cache_ttl, max_entries=cache_max_entries)

  def __repr__(self) -> str:
    return f'SDK(images={self._images}, engine={self._engine})'

  def authorize(self, token: str, tournId: str) -> bool:
    with Session(self._engine) as ses:
      results = ses.exec(select(Token).where(Token.token == token, Token.tournId == tournId))
      return results.first() is not None

  def tournaments(self) -> Sequence[Tournament]:
    with Session(self._engine) as ses:
      return ses.exec(select(Tournament)).all()

  def tournament(self, tournId: str) -> Tournament | None:
    with Session(self._engine) as ses:
      return ses.exec(queries.select_tnmt(tournId)).first()
    
  def group(self, tournId: str, group: str) -> Group | None:
    with Session(self._engine) as ses:
      return ses.exec(queries.select_group(tournId, group)).first()
  
  def round(self, **roundId: Unpack[cp.RoundId]) -> Sequence[FrontendGame]:
    with Session(self._engine) as ses:
      order: Any = Game.index # order_by's typing is messed up
      stmt = select(Game).where(*queries.round_games(**roundId)).order_by(order)
      return [FrontendGame.of(g) for g in ses.exec(stmt).all()]
    
  def _round_pgn(self, roundId: tuple[str, str, str]):
    rid = cp.roundId(*roundId)
    with Session(self._engine) as ses:
      order: Any = Game.index
      round = ses.exec(queries.select_round(**rid)).first()
      stmt = select(Game).where(*queries.round_games(**rid)).order_by(order)
      games = ses.exec(stmt).all()
      tnmt = ses.exec(queries.select_tnmt(rid['tournId'])).first()
      yield from export_all(games, tnmt, round)
    
  def round_pgn(self, *, skip: int = 0, take: int | None = None, tournId: str, group: str, round: str, no_cache: bool = False) -> Iterable[str]:
    id = (tournId, group, round)
    iter = self.pgns_cache[id] if not no_cache else self.pgns_cache.insert(id)
    iter = Iter(iter).skip(skip)
    return iter.take(take) if take is not None else iter  

  def group_pgn(self, *, skip: int = 0, take: int | None = None, no_cache: bool = False, **groupId: Unpack[cp.GroupId]) -> Iterable[str]:
    with Session(self._engine) as ses:
      group = ses.exec(queries.select_group(**groupId)).first()
      if group:
        for round in group.rounds:
          yield from self.round_pgn(skip=skip, take=take, **groupId, round=round, no_cache=no_cache)

  def game_pgn(self, **gameId: Unpack[cp.GameId]) -> FrontendPGN | None:
    with Session(self._engine) as ses:
      game = ses.exec(queries.select_game(**gameId)).first()
      if game and game.pgn:
        return game.pgn
      
  def images(self, **gameId: Unpack[cp.GameId]) -> Sequence[str] | None:
    with Session(self._engine) as ses:
      game = ses.exec(queries.select_game(**gameId)).first()
      if game:
        return [img.descaled_url for img in game.imgs]
  
  async def _upload(self, image: UploadFile, url: str):
    return await self._images.insert(url, await image.read())
  
  async def _descale_imgs(self, imgs: Sequence[tuple[str, bytes]], target_height: int):
    """Descales and uploads the images. Expects `(url, img)` tuples"""
    descaled = [descale2jpg(img, target_height) for _, img in imgs]
    uploads = [self._images.insert(url, img) for (url, _), img in zip(imgs, descaled)]
    return await asyncio.gather(*uploads)

  async def post_game(
    self, images: Sequence[UploadFile],
    descaled_height: int = 768, **gameId: Unpack[cp.GameId]
  ) -> tuple[Either, Coroutine]:
    """ - Runs tasks that must happen before responding to the client.
    - Returns the result + a coroutine that must run after responding (independent of the result)
    """
    img_urls = []
    descaled_urls = []
    delete_urls = []
    imgs = []

    try:
      with Session(self._engine) as ses:
        game = ses.exec(queries.select_game(**gameId)).first()
        if game is None:
          return Left(InexistentItem(detail=f'Game {gameId} not found in DB')), P.of(None).run()
        
        uuid = f'{stringify(**gameId)}_{uuid4()}'
        img_urls = [f'{uuid}/{i}.' + (safe_extension(img) or 'jpg') for i, img in enumerate(images)]
        descaled_urls = [f'{uuid}/{i}-{descaled_height}.jpg' for i in range(len(images))]

        imgs = await asyncio.gather(*[img.read() for img in images])
        original_uploads = [self._images.insert(url, img) for url, img in zip(img_urls, imgs)]
        res = E.sequence(await asyncio.gather(*original_uploads))

        delete_urls = [img.url for img in game.imgs] + [img.descaled_url for img in game.imgs]
        [ses.delete(img) for img in game.imgs]
        game.imgs = [Image(url=url, descaled_url=url) for url in img_urls]
        game.status = 'uploaded'
        ses.add(game)
        ses.commit()
    except Exception as e:
      res = Left(e)

    if res.tag == 'right':
      async def post_tasks():
        deletions = [self._images.delete(url) for url in delete_urls]
        descaled_uploads = self._descale_imgs(list(zip(descaled_urls, imgs)), descaled_height)
        await asyncio.gather(*deletions, descaled_uploads)

        with Session(self._engine) as ses:
          if (game := ses.exec(queries.select_game(**gameId)).first()):
            for img, descaled_url in zip(game.imgs, descaled_urls):
              img.descaled_url = descaled_url
              ses.add(img)
            ses.commit()

    else:
      async def post_tasks():
        await asyncio.gather(*[self._images.delete(url) for url in img_urls]) # not to leak storage

    return res, post_tasks()



from typing import Sequence
from sqlmodel import Session, select
from dslog import Logger
from moveread.dfy import Tournament, Group, SheetModel, queries, Token

def upsert_tnmt(session: Session, tnmt: Tournament, *, logger: Logger = Logger.click()):
  if (curr := session.exec(queries.select_tnmt(tnmt.tournId)).first()):
    logger('Deleting existing entry:', tnmt)
    session.delete(curr)
  logger('Inserting new entry:', tnmt)
  session.add(tnmt)

def upsert_group(
  session: Session, tournId: str, group: str, *,
  logger: Logger = Logger.click(), rounds: Sequence[str] = []
):
  if (grp := session.exec(queries.select_group(tournId, group)).first()):
    logger('Deleting existing groups:', group)
    session.delete(grp)
  
  grp = Group(tournId=tournId, name=group, rounds=rounds)
  logger('Inserting new group:', grp)
  session.add(grp)

def upsert_model(session: Session, tournId: str, model: str, *, logger: Logger = Logger.click()):
  if (curr := session.exec(queries.select_model(tournId)).first()):
    if curr.model == model:
      logger(f'Model "{model}" already set for "{tournId}"')
    else:
      logger(f'Updating model for "{tournId}": "{curr.model}" -> "{model}"')
      curr.model = model
      session.add(curr)
  else:
    logger(f'Inserting new model for "{tournId}": "{model}"')
    session.add(SheetModel(tournId=tournId, model=model))

def list_tokens(session: Session, tournId: str):
  stmt = select(Token.token).where(Token.tournId == tournId)
  return session.exec(stmt).all()

def add_token(session: Session, tournId: str, token: str, *, logger: Logger = Logger.click()):
  if session.exec(select(Token).where(Token.tournId == tournId, Token.token == token)).first():
    logger(f'Token "{token}" already exists for "{tournId}"')
  else:
    logger(f'Adding token "{token}" for "{tournId}"')
    session.add(Token(tournId=tournId, token=token))

def revoke_token(session: Session, tournId: str, token: str, *, logger: Logger = Logger.click()):
  stmt = select(Token).where(Token.tournId == tournId, Token.token == token)
  if (tok := session.exec(stmt).first()):
    logger(f'Revoking token "{token}" for "{tournId}"')
    session.delete(tok)
  else:
    logger(f'Token "{token}" not found for "{tournId}"')

class API:
  def __init__(self, session: Session, logger: Logger = Logger.click()):
    self.session = session
    self.logger = logger

  def list_tnmts(self):
    return self.session.exec(select(Tournament.tournId)).all()

  def upsert_tnmt(self, tnmt: Tournament):
    upsert_tnmt(self.session, tnmt, logger=self.logger)

  def upsert_group(self, tournId: str, group: str, *, rounds: Sequence[str] = []):
    upsert_group(self.session, tournId, group, rounds=rounds, logger=self.logger)

  def upsert_model(self, tournId: str, model: str):
    upsert_model(self.session, tournId, model, logger=self.logger)

  def list_tokens(self, tournId: str):
    return list_tokens(self.session, tournId)
  
  def add_token(self, tournId: str, token: str):
    add_token(self.session, tournId, token, logger=self.logger)

  def revoke_token(self, tournId: str, token: str):
    revoke_token(self.session, tournId, token, logger=self.logger)



from typing import TypeVar, Generic, Callable, Iterable
from time import time

K = TypeVar('K')
T = TypeVar('T')

class TimedIterableCache(Generic[K, T]):
  def __init__(self, get: Callable[[K], Iterable[T]], *, ttl_secs: float, max_entries: int):
    self._get = get
    self.ttl_secs = ttl_secs
    self._cache: dict[K, tuple[list[T], float]] = {}
    self.max_entries = max_entries

  def insert(self, k: K, access_time: float | None = None) -> Iterable[T]:
    access_time = access_time or time()
    if len(self._cache) >= self.max_entries:
      older_client = min(self._cache.keys(), key=self._cache.__getitem__)
      del self._cache[older_client]
    
    def gen():
      stored = []
      for x in self._get(k):
        stored.append(x)
        yield x
      self._cache[k] = stored, access_time

    yield from gen()


  def __getitem__(self, key: K) -> Iterable[T]:
    now = time()
    if key not in self._cache or now - self._cache[key][1] > self.ttl_secs:
      yield from self.insert(key, now)
    else:
      yield from self._cache[key][0]


from .queries import group_games, round_games, exact_game, exact_round, \
  select_game, select_group_games, select_tnmt, select_group, select_round, \
  select_rounds, select_tnmt_groups, select_tnmt_rounds, select_model, \
  uploaded_games, group_rounds, select_round_games

__all__ = [
  'group_games', 'round_games', 'exact_game', 'exact_round',
  'select_game', 'select_group_games', 'select_tnmt', 'select_group', 'select_round',
  'select_rounds', 'select_tnmt_groups', 'select_tnmt_rounds', 'select_model',
  'uploaded_games', 'group_rounds', 'select_round_games',
]


from argparse import ArgumentParser
import os

DEFAULT_ORIGINS = [
  'http://localhost:5173',
  'http://localhost:4713',
  'https://moveread.com',
  'https://dfy.moveread.com',
]

def env(variable: str, *, default = None, required: bool = True) -> dict:
  if (value := os.getenv(variable, default)) is not None:
    return dict(default=value)
  return dict(required=required)

def main():
  parser = ArgumentParser()
  parser.add_argument('--blob', type=str, **env('BLOB_CONN_STR'), help='Blobs KV connection string  ')
  parser.add_argument('--sql', type=str, **env('SQL_CONN_STR'), help='SQL connection string')
  parser.add_argument('--images', type=str, help='Local images path')
  parser.add_argument('--ttl', type=float, default=60*5, help='PGN Cache TTL (seconds)')
  parser.add_argument('--max-entries', type=int, default=1000, help='PGN Cache max entries')

  parser.add_argument('-p', '--port', type=int, default=80)
  parser.add_argument('--host', type=str, default='0.0.0.0')
  parser.add_argument('--cors', default=DEFAULT_ORIGINS, nargs='*', type=str, help='CORS allowed origins')

  args = parser.parse_args()

  from dslog import Logger, formatters
  logger = Logger.stderr().format(formatters.click)
  logger('Starting API...')
  logger('- Cache TTL: ', args.ttl)
  logger('- Cache Max Entries: ', args.max_entries)
  logger('- CORS allowed origins: ', args.cors)

  from kv import KV, LocatableKV
  blobs = KV.of(args.blob)
  assert isinstance(blobs, LocatableKV)

  from sqlmodel import create_engine
  engine = create_engine(args.sql)

  from fastapi.middleware.cors import CORSMiddleware
  from moveread.dfy import server
  import uvicorn

  sdk = server.SDK(blobs, engine, cache_ttl=args.ttl, cache_max_entries=args.max_entries)
  api = server.fastapi(sdk, blobs=blobs, logger=logger, images_path=args.images)
  api.add_middleware(CORSMiddleware, allow_origins=args.cors, allow_credentials=True, allow_methods=["*"], allow_headers=["*"])

  uvicorn.run(api, host=args.host, port=args.port)


from typing import Sequence, Literal
import os
import asyncio
from datetime import datetime, timedelta
from fastapi import FastAPI, Response, Request, status, UploadFile
from fastapi.responses import StreamingResponse
from fastapi.staticfiles import StaticFiles
from dslog import Logger
from dslog.uvicorn import setup_loggers_lifespan, DEFAULT_FORMATTER, ACCESS_FORMATTER
from kv import LocatableKV
from chess_pairings import roundId, gameId, groupId, GameId
from .._types import FrontendGame, Tournament, Group, FrontendPGN
from .sdk import SDK

async def notify(gameId: GameId, logger: Logger):
  endpoint: str = os.getenv('NOTIFY_ENDPOINT') # type: ignore
  token = os.getenv('NOTIFY_TOKEN')
  try:
    if endpoint is None:
      raise ValueError('No NOTIFY_ENDPOINT environment variable')
    import aiohttp
    async with aiohttp.ClientSession() as session:
      async with session.post(endpoint, json=gameId, headers=[('Authorization', f'Bearer {token}')]) as r:
        if r.status == 200:
          logger(f'Notified "{endpoint}" of game {gameId}', level='DEBUG')
        else:
          logger(f'Failed to notify "{endpoint}" of game {gameId}: {r.status}. Content:', await r.text(), level='ERROR')
  except Exception as e:
    logger(f'Exception notifying "{endpoint}" of game {gameId}:', e, level='ERROR')

def fastapi(
  sdk: SDK, *, images_path: str | None = None,
  blobs: LocatableKV[bytes],
  logger = Logger.click().prefix('[DFY API]')
):
  
  app = FastAPI(
  generate_unique_id_function=lambda route: route.name,
    lifespan=setup_loggers_lifespan(
      access=logger.format(ACCESS_FORMATTER),
      uvicorn=logger.format(DEFAULT_FORMATTER),
    )
  )

  if images_path is not None:
    app.mount('/images', StaticFiles(directory=images_path))

  @app.get('/authorize/{tournId}', responses={ 401: {}, 200: {} })
  def authorize(token: str, tournId: str):
    authed = sdk.authorize(token, tournId)
    if not authed:
      logger(f'Unauthorized access to tournament {tournId}: {token}', level='DEBUG')
    return Response(status_code=200 if authed else status.HTTP_401_UNAUTHORIZED)

  @app.get('/')
  def tournaments() -> Sequence[Tournament]:
    return sdk.tournaments()

  @app.get('/version')
  def version():
    return '0.1.22'

  @app.get('/{tournId}', responses={ 404: {}, 200: { 'model': Tournament }})
  def tournament(tournId: str, r: Response) -> Tournament | None:
    tnmt = sdk.tournament(tournId)
    if tnmt is None:
      r.status_code = status.HTTP_404_NOT_FOUND
    return tnmt
  
  @app.get('/{tournId}/{group}.pgn', response_model=str)
  def group_pgn(tournId: str, group: str, skip: int = 0, take: int | None = None, no_cache: bool = False):
    pgn = sdk.group_pgn(**groupId(tournId, group), skip=skip, take=take, no_cache=no_cache)
    return StreamingResponse(content=pgn, media_type='application/x-chess-pgn', headers={
      'Content-Disposition': f'attachment; filename={tournId}_{group}.pgn'
    })
  
  @app.get('/{tournId}/{group}', responses={ 404: {}, 200: { 'model': Group }})
  def group(tournId: str, group: str, r: Response) -> Group | None:
    grp = sdk.group(tournId, group)
    if grp is None:
      r.status_code = status.HTTP_404_NOT_FOUND
    return grp

  @app.get('/{tournId}/{group}/{round}.pgn')
  def round_pgn(tournId: str, group: str, round: str, skip: int = 0, take: int | None = None, no_cache: bool = False):
    pgn = sdk.round_pgn(**roundId(tournId, group, round), skip=skip, take=take, no_cache=no_cache)
    return StreamingResponse(content=pgn, media_type='application/x-chess-pgn', headers={
      'Content-Disposition': f'attachment; filename={tournId}_{group}_{round}.pgn'
    })
  
  @app.get('/{tournId}/{group}/{round}')
  def round(tournId: str, group: str, round: str) -> Sequence[FrontendGame]:
    return sdk.round(**roundId(tournId, group, round))

  @app.get('/{tournId}/{group}/{round}/{board}/pgn', responses={ 404: {}, 200: { 'model': FrontendPGN }})
  def game_pgn(tournId: str, group: str, round: str, board: str, r: Response) -> FrontendPGN | None:
    pgn = sdk.game_pgn(**gameId(tournId, group, round, board))
    if pgn is None:
      r.status_code = status.HTTP_404_NOT_FOUND
    return pgn
  
  @app.get('/{tournId}/{group}/{round}/{board}/images', responses={ 404: {}, 200: { 'model': Sequence[str] }})
  def images(tournId: str, group: str, round: str, board: str, req: Request, res: Response) -> Sequence[str] | None:
    urls = sdk.images(**gameId(tournId, group, round, board))
    if urls is None or len(urls) == 0:
      res.status_code = status.HTTP_404_NOT_FOUND
      return None
    
    if images_path is not None:
      path = os.path.join(str(req.base_url), 'images')
      return [os.path.join(path, url) for url in urls]
    else:
      expiry = datetime.now() + timedelta(hours=1)
      return [blobs.url(url, expiry=expiry) for url in urls]
  
  @app.post('/{tournId}/{group}/{round}/{board}', responses={
    401: dict(model=Literal['UNAUTHORIZED']),
    200: dict(model=Literal['OK']),
    500: dict(model=Literal['ERROR'])
  })
  async def post_game(
    tournId: str, group: str, round: str, board: str, token: str,
    images: list[UploadFile], r: Response, 
  ) -> Literal['OK', 'UNAUTHORIZED', 'ERROR']:
  
    if not sdk.authorize(token, tournId):
      r.status_code = status.HTTP_401_UNAUTHORIZED
      return 'UNAUTHORIZED'
    
    gid = gameId(tournId, group, round, board)
    res, coro = await sdk.post_game(images, **gid)

    async def post_task():
      r = await coro
      logger('Post task finished, returning:', r)
    asyncio.create_task(post_task())
    
    if res.tag == 'left':
      logger(f'Error posting game {gid}:', res.value, level='ERROR')
      r.status_code = status.HTTP_500_INTERNAL_SERVER_ERROR
      return 'ERROR'
    
    asyncio.create_task(notify(gid, logger))
    return 'OK'  
  
  return app


from typing import Sequence, ClassVar, Literal
from dataclasses import dataclass
from datetime import date, datetime
from sqlmodel import Field, SQLModel, Relationship
from sqltypes import SpaceDelimitedList, ValidatedStr, ValidatedJSON
from chess_pairings import Paired, GameId, gameId, Result
from chess_scraping import Source

class SheetModel(SQLModel, table=True):
  tournId: str = Field(primary_key=True, foreign_key='tournament.tournId')
  model: str

class Tournament(SQLModel, table=True):
  tournId: str = Field(primary_key=True)
  name: str
  site: str | None = None
  start_date: date
  end_date: date
  groups: Sequence[str] = Field(sa_type=SpaceDelimitedList)

class Pairings(SQLModel, table=True):
  tournId: str = Field(foreign_key='tournament.tournId', primary_key=True)
  group: str = Field(primary_key=True)
  pairings: Source = Field(sa_type=ValidatedJSON(Source))

class Group(SQLModel, table=True):
  tournId: str = Field(primary_key=True, foreign_key='tournament.tournId')
  name: str = Field(primary_key=True)
  rounds: Sequence[str] = Field(sa_type=SpaceDelimitedList)

class Round(SQLModel, table=True):
  tournId: str = Field(primary_key=True, foreign_key='tournament.tournId')
  group: str = Field(primary_key=True)
  name: str = Field(primary_key=True)
  start_dtime: datetime
  
class Image(SQLModel, table=True):
  id: int | None = Field(default=None, primary_key=True)
  url: str
  descaled_url: str
  gameId: int = Field(default=None, foreign_key='game.id')

class FrontendPGN(SQLModel):
  moves: Sequence[str] = Field(sa_type=SpaceDelimitedList)
  early: bool | None = None

class PGN(FrontendPGN, table=True):
  gameId: int = Field(default=None, primary_key=True, foreign_key='game.id')

@dataclass
class FrontendGame:
  board: str
  pairing: Paired
  status: 'Game.Status | None' = None

  @classmethod
  def of(cls, game: 'Game'):
    return cls(game.board, Paired(white=game.white, black=game.black, result=game.result), game.status)

class Game(SQLModel, table=True):
  Status: ClassVar = Literal['uploaded', 'doing', 'done']
  status: 'Game.Status | None' = Field(default=None, sa_type=ValidatedStr(Status))
  id: int | None = Field(default=None, primary_key=True)
  tournId: str = Field(foreign_key='tournament.tournId')
  group: str
  round: str
  board: str
  index: int
  """Boards may have out of whack names (e.g. in team tournaments "1.3"). This is the order you'd see in chess-results"""
  white: str
  black: str
  result: Result | None = Field(sa_type=ValidatedStr(Result))
  imgs: list[Image] = Relationship()
  pgn: PGN | None = Relationship()

  def gameId(self) -> GameId:
    return gameId(self.tournId, self.group, self.round, self.board)
  
class Token(SQLModel, table=True):
  id: int | None = Field(default=None, primary_key=True)
  token: str
  tournId: str = Field(foreign_key='tournament.tournId')
  
__all__ = [
  'Game', 'GameId', 'Image', 'Tournament', 'PGN', 'Pairing', 'Token'
]